@@ -13,10 +13,11 @@
 
 int main(int argc,char *argv[])
 { Scsi_Command	cmd;
-  unsigned char	header[36],inq[128],*page2A=NULL;
-  char          cmdname[64];
-  int	        i,j,ntracks,err,dvd_minus=0,dvd_plus=0,plus_mediaid_printed=0;
-  unsigned short profile;
+  unsigned char	header[40],inq[128],*page2A=NULL;
+  char		cmdname[64];
+  int		i,j,ntracks,err,dvd_dash=0,dvd_plus=0,
+		plus_mediaid_printed=0,page2A_printed=0,dvd_ram_spare=0;
+  unsigned short profile,dvd_0E=0,dvd_10=0,dvd_11=0,dvd_0A=0,dvd_C0=0;
 
     if (argc<2)
 	fprintf (stderr,"usage: %s /dev/dvd\n",argv[0]),
@@ -97,6 +98,7 @@ int main(int argc,char *argv[])
 	for(i=8+(pages[6]<<8|pages[7]);i<len;)
 	{   if (pages[i] == 0)	break;
 
+	    if (pages[i] == 0x2A) page2A_printed=1;
 	    printf (" %02X:",pages[i++]);
 	    n=pages[i++];
 	    for (j=0;j<n;j++)
@@ -123,7 +125,7 @@ int main(int argc,char *argv[])
 
     { int len,n;
 
-	page2A=header, len=sizeof(header);
+	page2A=header, len=36;
 
 	while (1)
 	{   cmd[0] = 0x5A;	// MODE SENSE(10)
@@ -151,7 +153,7 @@ int main(int argc,char *argv[])
 		exit (1);
 
 	    if (page2A==header) page2A=NULL;
-	    else if (argc>2)
+	    else if (argc>2 && !page2A_printed)
 	    {	printf("MODE SENSE[#2A]:\n");
 		for(i=8+(page2A[6]<<8|page2A[7]);i<len;)
 		{   printf (" %02X:",page2A[i++]);
@@ -168,6 +170,48 @@ int main(int argc,char *argv[])
 	}
     }
 
+    do
+    { unsigned int   len,j;
+      unsigned char *p;
+
+	cmd[0] = 0xAD;
+	cmd[7] = 0xFF;
+	cmd[9] = 4;
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,header,4)))
+	{   if (argc>2)
+		sperror ("READ DVD STRUCTURE#FF",err);
+	    break;
+	}
+
+	len = (header[0]<<8|header[1]) + 2;
+	if (!(p = (unsigned char *)malloc(len))) break;
+
+	cmd[0] = 0xAD;
+	cmd[7] = 0xFF;
+	cmd[8] = len>>8;
+	cmd[9] = len;
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,p,len)))
+	    break;
+
+	for (j=4;j<len;j+=4)
+	    switch (p[j])
+	    {	case 0x0A:  dvd_0A=p[j+2]<<8|p[j+3];	break;
+		case 0x0E:  dvd_0E=p[j+2]<<8|p[j+3];	break;
+		case 0x10:  dvd_10=p[j+2]<<8|p[j+3];	break;
+		case 0x11:  dvd_11=p[j+2]<<8|p[j+3];	break;
+		case 0xC0:  dvd_C0=p[j+2]<<8|p[j+3];	break;
+	    }
+
+	if (argc<=2) break;
+
+	printf ("READ DVD STRUCTURE[#FF]:\n");
+	for (j=4;j<len;j+=4)
+	printf(" STRUCTURE#%02x           %02x:%d\n",p[j],p[j+1],p[j+2]<<8|p[j+3]);
+
+    } while (0);
+
     cmd[0] = 0x46;	// GET CONFIGURATION
     cmd[1] = 1;
     cmd[8] = 8;
@@ -218,7 +262,7 @@ int main(int argc,char *argv[])
     if (profile==0)
 	fprintf (stderr,":-( no media mounted, exiting...\n"),
 	exit(1);
-    if ((profile&0xF0) != 0x10)
+    if ((profile&0xF0) != 0x10 && profile != 0x2B)
 	fprintf (stderr,":-( non-DVD media mounted, exiting...\n"),
 	exit (1);
 
@@ -232,17 +276,18 @@ int main(int argc,char *argv[])
 	    case 0x14:	str="-RW Sequential";		break;
 	    case 0x1A:	str="+RW";			break;
 	    case 0x1B:	str="+R";			break;
+	    case 0x2B:	str="+R Double Layer";		break;
 	    default:	str="unknown";			break;
 	}
 	printf (" Mounted Media:         %Xh, DVD%s\n",profile,str);
-	if (str[0]=='-' && str[3]!='M')	dvd_minus=0x10;
+	if (str[0]=='-' && str[3]!='M')	dvd_dash=0x10;
 	if (str[0]=='+')		dvd_plus=0x11;
     }
 
-    do
+    if (dvd_0E || dvd_11) do
     { union { unsigned char _e[4+40],_11[4+256]; } dvd;
-      unsigned int len;
-      unsigned char format=dvd_minus?0xE:0x11;
+      unsigned int  len;
+      unsigned char format=dvd_dash?0x0E:0x11;
 
 	cmd[0] = 0xAD;
 	cmd[7] = format;
@@ -268,15 +313,15 @@ int main(int argc,char *argv[])
 
 	if (argc>2)
 	{   printf ("READ DVD STRUCTURE[#%02xh]:",format);
-	    for (j=0,i=dvd_minus?sizeof(dvd._e):sizeof(dvd._11);j<i;j++)
+	    for (j=0,i=dvd_dash?sizeof(dvd._e):sizeof(dvd._11);j<i;j++)
 		printf("%c%02x",j?' ':'\t',dvd._11[j]);
 	    printf("\n");
 	}
 
-	if (dvd_minus && dvd._e[4+16]==3 && dvd._e[4+24]==4)
+	if (dvd_dash && dvd_0E && dvd._e[4+16]==3 && dvd._e[4+24]==4)
 	printf (" Media ID:              %6.6s%-6.6s\n",dvd._e+4+17,dvd._e+4+25);
 
-	if (!dvd_minus)
+	if (dvd_plus && dvd_11)
 	printf (" Media ID:              %.8s/%.3s\n",dvd._11+23,dvd._11+31),
 	plus_mediaid_printed=1;
 
@@ -296,11 +341,11 @@ int main(int argc,char *argv[])
 	// Check Write Speed descriptors for sanity. Some DVD+units
 	// return CD-R descriptors here, which has no meaning in the
 	// context of interest.
-	if (v<1385)	break;
+	if (v<1352)	break;
 	if (len>=(hlen+32))
 	{   n = (p[30]<<8|p[31])*4;
 	    for (i=0;i<n;i+=4)
-		if ((p[32+i+2]<<8|p[32+i+3]) < 1385) break;
+		if ((p[32+i+2]<<8|p[32+i+3]) < 1352) break;
 	    if (i<n) break;
 	}
 
@@ -369,9 +414,7 @@ int main(int argc,char *argv[])
 	}
 
 	if ((p[4]&2) == 0)
-	{   fprintf (stderr,":-( unit fails to report current write performance\n");
-	    break;
-	}
+	    fprintf (stderr,":-( reported write performance might be bogus\n");
 
 	if (argc<=2) printf ("GET [CURRENT] PERFORMANCE:\n");
 
@@ -453,53 +496,23 @@ int main(int argc,char *argv[])
 
 legacy:
 
-    if (argc>2) do
-    { unsigned int len,j;
-      unsigned char *p;
-
-	cmd[0] = 0xAD;
-	cmd[7] = 0xFF;
-	cmd[9] = 4;
-	cmd[11] = 0;
-	if ((err=cmd.transport(READ,header,4)))
-	{   if (argc>2)
-		sperror ("READ DVD STRUCTURE#FF",err);
-	    break;
-	}
-
-	len = (header[0]<<8|header[1]) + 2;
-	if (!(p = (unsigned char *)malloc(len))) break;
-
-	cmd[0] = 0xAD;
-	cmd[7] = 0xFF;
-	cmd[8] = len>>8;
-	cmd[9] = len;
-	cmd[11] = 0;
-	if ((err=cmd.transport(READ,p,len)))
-	    break;
-
-	printf ("READ DVD STRUCTURE[#FF]:\n");
-	for (j=4;j<len;j+=4)
-	printf(" STRUCTURE#%02x           %02x:%d\n",p[j],p[j+1],p[j+2]<<8|p[j+3]);
-
-    } while (0);
-
     header[4]=0;
     cmd[0] = 0xAD;	// READ DVD STRUCTURE
-    cmd[7] = dvd_minus;
+    cmd[7] = dvd_dash;
     cmd[9] = 36;
     cmd[11] = 0;
     if ((err=cmd.transport(READ,header,36)))
-    {   if (dvd_minus) { dvd_minus=0; goto legacy; }
-	sprintf (cmdname,"READ DVD STRUCTURE#%X",dvd_minus),
-	sperror (cmdname,err);
+    {	if (dvd_dash) { dvd_dash=0; goto legacy; }
+	if (err!=0x52400 || argc>2)
+	    sprintf (cmdname,"READ DVD STRUCTURE#%X",dvd_dash),
+	    sperror (cmdname,err);
     }
     else do
     { unsigned char book=header[4];
       const char *brand;
       unsigned int phys_start,phys_end;
 
-	printf("READ DVD STRUCTURE[#%Xh]:",dvd_minus<0?0:dvd_minus);
+	printf("READ DVD STRUCTURE[#%Xh]:",dvd_dash<0?0:dvd_dash);
 	if (argc>2)
 	    for (j=0;j<20-4;j++) printf("%c%02x",j?' ':'\t',header[4+j]);
 	printf("\n");
@@ -510,6 +523,7 @@ legacy:
 	    case 0x30:	brand="-RW";	break;
 	    case 0x90:	brand="+RW";	break;
 	    case 0xA0:	brand="+R";	break;
+	    case 0xE0:	brand="+R DL";	break;
 	    default:	brand=NULL;	break;
 	}
 	printf (" Media Book Type:       %02Xh, ",book);
@@ -517,26 +531,34 @@ legacy:
 			    brand,book&0xF);
 	else	    printf ("unrecognized value\n");
 
+	if (header[4+1]&0xF0)	dvd_ram_spare=5120;
+	else			dvd_ram_spare=12800;
+
 	if (dvd_plus && !plus_mediaid_printed)
 	printf (" Media ID:              %.8s/%.3s\n",header+23,header+31);
 
-	phys_start  = header[4+5]<<16;
-	phys_start |= header[4+6]<<8;
+	phys_start  = header[4+5]<<16,
+	phys_start |= header[4+6]<<8,
 	phys_start |= header[4+7];
-	phys_end    = header[4+9]<<16;
-	phys_end   |= header[4+10]<<8;
-	phys_end   |= header[4+11];
-	if (phys_end>0)	phys_end-=phys_start;
+	if ((header[4+2]&0x60)==0)	// single-layer
+	    phys_end  = header[4+9]<<16,
+	    phys_end |= header[4+10]<<8,
+	    phys_end |= header[4+11];
+	else
+	    phys_end  = header[4+13]<<16,
+	    phys_end |= header[4+14]<<8,
+	    phys_end |= header[4+15];
+	if (phys_end>0)	phys_end -= phys_start;
 	if (phys_end>0) phys_end += 1;
 
 	printf (" %s    %u*2KB=%llu\n",
-		dvd_minus>=0?"Legacy lead-out at:":"Last border-out at:",
+		dvd_dash>=0?"Legacy lead-out at:":"Last border-out at:",
 		phys_end,phys_end*2048LL);
 
-	if (dvd_minus<=0) break;
+	if (dvd_dash<=0) break;
 
 	cmd[0] = 0xAD;
-	cmd[7] = 0; dvd_minus=-1;
+	cmd[7] = 0; dvd_dash=-1;
 	cmd[9] = 20;
 	cmd[11] = 0;
 	if ((err=cmd.transport(READ,header,20)))
@@ -549,6 +571,80 @@ legacy:
 	printf ("DVD-ROM media detected, exiting...\n"),
 	exit(0);
 
+    if (profile==0x2B) do
+    { unsigned char s[4+8];
+
+	cmd[0] = 0xAD;	// READ DVD STRUCTURE
+	cmd[7] = 0x20;	// "DVD+R Double Layer Boundary Information"
+	cmd[9] = sizeof(s);
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,s,sizeof(s))))
+	{   sperror ("READ DVD+R DL BOUNDARY INFORMATION",err);
+	    break;
+	}
+
+	if ((s[0]<<8|s[1]) < 10)
+	{   fprintf (stderr,":-( insane DVD+R DL BI structure length\n");
+	    break;
+	}
+
+	printf ("DVD+R DOUBLE LAYER BOUNDARY INFORMATION:\n");
+	printf (" L0 Data Zone Capacity: %u*2KB, can %s be set\n",
+					 s[8]<<24|s[9]<<16|s[10]<<8|s[11],
+					 s[4]&0x80?"no longer":"still");
+    } while (0);
+
+    if (profile==0x12 && dvd_0A) do
+    { unsigned char s[16];
+      unsigned int  a,b;
+
+	if (dvd_0A>sizeof(s))	dvd_0A=sizeof(s);
+
+	cmd[0] = 0xAD;	// READ DVD STRUCTURE
+	cmd[7] = 0x0A;	// "DVD-RAM Spare Area Information"
+	cmd[8] = dvd_0A>>8;
+	cmd[9] = dvd_0A;
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,s,dvd_0A)))
+	{   sperror ("READ DVD-RAM SPARE INFORMATION",err);
+	    break;
+	}
+
+	if (dvd_0A<8)	break;
+	printf ("DVD-RAM SPARE AREA INFORMATION:\n");
+	a = s[ 4]<<24|s[ 5]<<16|s[ 6]<<8|s[ 7];
+	b = dvd_ram_spare;
+	printf (" Primary SA:            %u/%u=%.1f%% free\n",a,b,(a*100.0)/b);
+	if (dvd_0A<16)	break;
+	a = s[ 8]<<24|s[ 9]<<16|s[10]<<8|s[11];
+	b = s[12]<<24|s[13]<<16|s[14]<<8|s[15];
+	printf (" Supplementary SA:      %u/%u=%.1f%% free\n",a,b,(a*100.0)/b);
+    } while (0);
+
+    if (profile==0x12 && dvd_C0) do
+    { unsigned char s[8];
+
+	if (dvd_C0>sizeof(s))	dvd_C0=sizeof(s);
+
+	cmd[0] = 0xAD;	// READ DVD STRUCTURE
+	cmd[7] = 0xC0;	// "Write Protection Status"
+	cmd[8] = dvd_C0>>8;
+	cmd[9] = dvd_C0;
+	cmd[11] = 0;
+	if ((err=cmd.transport(READ,s,dvd_C0)))
+	{   sperror ("READ WRITE PROTECTION STATUS",err);
+	    break;
+	}
+
+	if (dvd_C0<8)	break;
+	printf ("DVD-RAM WRITE PROTECTION STATUS:\n");
+	if (s[4]&0x04)
+	    printf (" Write protect tab on cartridge is set to protected\n");
+	if (s[4]&0x01)
+	    printf (" Software Write Protection until Power down is on\n");
+	printf (" Persistent Write Protection is %s\n",s[4]&0x02?"on":"off");
+    } while (0);
+
     cmd[0] = 0x51;	// READ DISC INFORMATION
     cmd[8] = 32;
     cmd[9] = 0;
@@ -587,7 +683,7 @@ legacy:
 	}
 	printf ("\n");
 
-	while (header[2]&0x10 && argc>2)
+	while (header[2]&0x10 && (profile==0x12 || argc>2))
 	{ unsigned char formats[260];
 	  int len;
 	  unsigned int capacity,blocksize;
@@ -640,34 +736,60 @@ legacy:
 	}
     }
 
+    if (argc<=2 && profile==0x12) ntracks=0;
+
     for (i=1;i<=ntracks;i++)
-    { static const char *track_state[]={
+    { static const char *dash_track_state[]={
 				"complete",	"complete incremental",
 				"blank",	"invisible incremental",
 				"partial",	"partial incremental",
 				"reserved",	"reserved incremental"};
+      static const char *plus_track_state[]={
+				"invisible",		"blank",
+				"partial/complete",	"reserved" };
+      int len=32;
 
-	cmd[0] = 0x52;	// READ TRACK INFORMATION
-	cmd[1] = 1;
-	cmd[2] = i>>24;
-	cmd[3] = i>>16;
-	cmd[4] = i>>8;
-	cmd[5] = i;
-	cmd[8] = 32;
-	cmd[9] = 0;
-	if ((err=cmd.transport(READ,header,32)))
-	    sperror ("READ TRACK INFORMATION",err),
-	    exit (1);
+	while (len)
+	{   cmd[0] = 0x52;	// READ TRACK INFORMATION
+	    cmd[1] = 1;
+	    cmd[2] = i>>24;
+	    cmd[3] = i>>16;
+	    cmd[4] = i>>8;
+	    cmd[5] = i;
+	    cmd[8] = len;
+	    cmd[9] = 0;
+	    if ((err=cmd.transport(READ,header,len)))
+	    {	if (i<ntracks) { i=ntracks; continue; }
+		sperror ("READ TRACK INFORMATION",err),
+		exit (1);
+	    }
+
+	    if ((profile==0x1B || profile==0x2B) &&
+		(header[6]&0x80)==0 && (header[0]<<8|header[1])>=38)
+	    {	if (len==32)	{ len=40; continue; }
+		else		{ break;            }
+	    }
+
+	    len = 0;
+	}
 
 	printf ("READ TRACK INFORMATION[#%d]:",i);
 	if (argc>2)
 	    for (j=0;j<16;j++) printf("%c%02x",j?' ':'\t',header[j]);
 	printf ("\n");
+	if ((profile&0x0F)==0x0B)	// DVD+R
+	{
+	printf (" Track State:           %s\n",
+					 plus_track_state[header[6]>>6]);
+	}
+	else
+	{
 	printf (" Track State:           %s%s",
 					 ((header[6]>>5)==1 && (header[7]&1))?"in":"",
-					 track_state[header[6]>>5]);
+					 dash_track_state[header[6]>>5]);
 	if (header[5]&0x20)	printf(",damaged");
 	printf ("\n");
+	}
 	printf (" Track Start Address:   %u*2KB\n",
 		header[8]<<24|header[9]<<16|header[10]<<8|header[11]);
 	if (header[7]&1)
@@ -683,6 +805,9 @@ legacy:
 	if (header[7]&2)
 	printf (" Last Recorded Address: %u*2KB\n",
 		header[28]<<24|header[29]<<16|header[30]<<8|header[31]);
+	if (len>=40)
+	printf (" ROM Compatibility LBA: %u\n",
+		header[36]<<24|header[37]<<16|header[38]<<8|header[39]);
     }
 
     do
@@ -750,4 +875,22 @@ legacy:
 	printf (" SONY Session Info:     #%u@%u\n",header[4+2],sony);
 
     } while (0);
+
+    do {
+	unsigned int ccity,bsize;
+	cmd[0] = 0x25;	// READ CAPACITY
+	cmd[9] = 0;
+	if ((err=cmd.transport (READ,header,8)))
+	{   if (argc>2)
+		sperror ("READ CAPACITY",err);
+	    break;
+	}
+
+	ccity = header[0]<<24|header[1]<<16|header[2]<<8|header[3], ccity++;
+	bsize = header[4]<<24|header[5]<<16|header[6]<<8|header[7];
+
+	printf ("READ CAPACITY:          %u*%u=%llu\n",
+					ccity,bsize,
+					(unsigned long long)ccity*bsize);
+    } while (0);
 }
