@@ -1,5 +1,5 @@
 /*
- * growisofs 5.19 by Andy Polyakov <appro@fy.chalmers.se>.
+ * growisofs 5.21 by Andy Polyakov <appro@fy.chalmers.se>.
  *
  * Use-it-on-your-own-risk, GPL bless...
  *
@@ -291,12 +291,36 @@
  * - average write speed report at the end of recording;
  * - LG GSA-4081B fails to "SET STREAMING" with "LBA OUT OF RANGE" for
  *   DVD+RW media, but not e.g. DVD-R;
+ * 5.20:
+ * - DVD-RAM reload if recorded with -poor-man;
+ * - -use-the-force-luke=wrvfy for WRITE AND VERIFY(10);
+ * - "flushing cache" takes forever, from 5.19-1;
+ * - HP-UX: inconsistency between /dev/rdsk and /dev/rscsi names;
+ * - handle non-fatal OPC errors;
+ * - DVD+R Double Layer support;
+ * - -use-the-force-luke=4gms to allow ISO9660 directory structures
+ *   to cross 4GB boundary, the option is effective only with DVD+R DL
+ *   and for data to be accessible under Linux isofs a kernel patch is
+ *   required;
+ * - more sane sanity check for -use-the-force-luke=tracksize:N;
+ * - -use-the-force-luke=break:size to set Layer Break position for
+ *   Double Layer recordings;
+ * - speed verification issue with 8x AccessTek derivatives addressed;
+ * - -use-the-force-luke=noload to leave tray ejected at the end;
+ * - allow to resume incomplete sessions recorded with -M option;
+ * - Layer Break position sanity check with respect to dataset size;
+ * - #if directive to get rid of sudo check at compile time with
+ *   'make WARN=-DI_KNOW_ALL_ABOUT_SUDO';
+ * 5.21:
+ * - Linux: fix for kernel version 2.6>=8, 2.6.8 itself is deficient,
+ *   but the problem can be worked around by installing this version
+ *   set-root-uid;
  */
 #define PRINT_VERSION(cmd)	do {			\
     char *s=strrchr((cmd),'/');				\
     s ? s++ : (s=(cmd));				\
     printf ("* %.*sgrowisofs by <appro@fy.chalmers.se>,"\
-	    " version 5.19,\n",(int)(s-(cmd)),(cmd));	\
+	    " version 5.21,\n",(int)(s-(cmd)),(cmd));	\
 } while (0)
 
 #define _LARGEFILE_SOURCE 
@@ -414,9 +438,10 @@ static void   *ioctl_handle=INVALID_HANDLE;
 #define ioctl_fd ((long)ioctl_handle)
 
 static int	poor_man=-1, zero_image=0, quiet=1,
-		overburn=0, no_tty_check=0, dry_run=0, dao_size=0;
+		overburn=0, no_tty_check=0, dry_run=0, dao_size=0,
+		no_4gb_check=0, layer_break=0;
 
-int    dvd_compat=0, test_write=0, no_reload=0, mmc_profile=0;
+int    dvd_compat=0, test_write=0, no_reload=0, mmc_profile=0, wrvfy=0;
 double speed_factor=0.0;
 char  *ioctl_device;
 int    _argc;
@@ -427,9 +452,16 @@ char **_argv;
 #include <linux/cdrom.h>
 #include <sys/ioctl.h>
 #include <linux/raw.h>
+#include <sys/prctl.h>
+#include <sys/syscall.h>
+#ifndef _LINUX_WAIT_H
+#define _LINUX_WAIT_H	/* linux headers are impaired */
+#endif
+#include <linux/capability.h>
 
 char *find_raw_device(struct stat64 *sb)
-{ int   i,rawctl,dev_major,dev_minor;
+{ int   i,rawctl;
+  dev_t dev_major,dev_minor;
   char  *ret=NULL;
   struct raw_config_request req;
   static char rawdevname[24] = "";
@@ -459,6 +491,7 @@ char *setup_fds (char *device)
 { char *odevice;
   uid_t uid=getuid();
   struct stat64 sb,sc;
+  int fd;
 
     /*
      * I ignore return values from set{re}uid calls because if
@@ -521,7 +554,17 @@ char *setup_fds (char *device)
 	return device;
     }
 
-    ioctl_handle=(void *)(long)dup(in_fd);
+    /*
+     * O_RDWR is expected to provide for none set-root-uid
+     * execution under Linux kernel 2.6[.8]. If it fails
+     * [as under 2.6.8], CAP_SYS_RAWIO in combination
+     * with set-root-uid should fill in for the kernel
+     * deficiency...
+     */
+    if ((fd = open64 (device,O_RDWR|O_NONBLOCK)) >= 0)
+	ioctl_handle=(void *)(long)fd;
+    else
+	ioctl_handle=(void *)(long)dup(in_fd);
 
     /*
      * get_mmc_profile terminates the program if ioctl_handle is
@@ -541,9 +584,33 @@ char *setup_fds (char *device)
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
 	case 0x1B:	/* DVD+R			*/
+	case 0x2B:	/* DVD+R Double Layer		*/
 	open_poor_man:
 	    poor_man = 1;
-	    out_fd = dup(in_fd);
+	    out_fd = dup(ioctl_fd);
+#if defined(PR_SET_KEEPCAPS)
+	    if (prctl(PR_SET_KEEPCAPS,1) == 0) do {
+#if defined(_LINUX_CAPABILITY_VERSION) && defined(CAP_SYS_RAWIO) && defined(SYS_capset)
+		struct __user_cap_header_struct	h;
+		struct __user_cap_data_struct	d;
+
+		h.version	= _LINUX_CAPABILITY_VERSION;
+		h.pid		= 0;
+		d.effective	= 0;
+		d.permitted	= 1<<CAP_SYS_RAWIO;
+		d.inheritable	= 0;
+		if (syscall(SYS_capset,&h,&d) < 0) break;
+
+		/* drop all privileges & effective capabilities */
+		/* setuid(uid) seems to preserve saved uid
+		   after dropping CAP_SETUID above... */
+		setresuid (uid,uid,uid);
+
+		d.effective	= 1<<CAP_SYS_RAWIO;
+		if (syscall(SYS_capset,&h,&d) < 0) break;
+#endif
+	    } while (0);
+#endif
 	    setuid (uid);	/* drop all privileges	*/
 	    return ioctl_device=device;
 	case 0x1A:	/* DVD+RW			*/
@@ -646,8 +713,9 @@ char *setup_fds (char *device)
     {	case 0x11:	/* DVD-R			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
-	case 0x1B:	/* DVD+R			*/
 	case 0x1A:	/* DVD+RW			*/
+	case 0x1B:	/* DVD+R			*/
+	case 0x2B:	/* DVD+R Double Layer		*/
 	open_poor_man:
 	    poor_man = 1;
 	    out_fd = dup(ioctl_fd);
@@ -757,8 +825,9 @@ char *setup_fds (char *device)
     {	case 0x11:	/* DVD-R			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
-	case 0x1B:	/* DVD+R			*/
 	case 0x1A:	/* DVD+RW			*/
+	case 0x1B:	/* DVD+R			*/
+	case 0x2B:	/* DVD+R Double Layer		*/
 	open_poor_man:
 	    poor_man = 1;
 	    out_fd = dup(in_fd);	/* it's ignored in poor_man ... */
@@ -863,8 +932,9 @@ char *setup_fds (char *device)
 	case 0x12:	/* DVD-RAM			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
-	case 0x1B:	/* DVD+R			*/
 	case 0x1A:	/* DVD+RW			*/
+	case 0x1B:	/* DVD+R			*/
+	case 0x2B:	/* DVD+R Double Layer		*/
 	    poor_man = 1;
 	    out_fd = dup(ioctl_fd);
 #if 0	    /* 'man uscsi' maintains that root privileges are required upon
@@ -1002,7 +1072,7 @@ char *setup_fds (char *device)
 #else
     if (major(sb.st_rdev) == SCTL_MAJOR)
 #endif
-	fprintf (stderr,":-( stick to /dev/rdsk/c%xt%x%c%x!\n",
+	fprintf (stderr,":-( stick to /dev/rdsk/c%ut%u%c%x!\n",
 			(m>>16)&0xFF,(m>>12)&0xF,'d',(m>>8)&0xF),
 	exit(FATAL_START(EINVAL));
 
@@ -1011,7 +1081,7 @@ char *setup_fds (char *device)
      * use /dev/rscsi/cXtYlZ for pass-through access in order to avoid
      * command replay by upper "class" driver...
      */
-    sprintf (rscsi,"/dev/rscsi/c%xt%x%c%x",
+    sprintf (rscsi,"/dev/rscsi/c%ut%u%c%x",
 		    (m>>16)&0xFF,(m>>12)&0xF,'l',(m>>8)&0xF);
     ioctl_handle = (void *)(long)open64 (rscsi,O_RDONLY);
     if (ioctl_handle == INVALID_HANDLE)
@@ -1038,8 +1108,9 @@ char *setup_fds (char *device)
 	case 0x12:	/* DVD-RAM			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
-	case 0x1B:	/* DVD+R			*/
 	case 0x1A:	/* DVD+RW			*/
+	case 0x1B:	/* DVD+R			*/
+	case 0x2B:	/* DVD+R Double Layer		*/
 	    poor_man = 1;
 	    out_fd = dup(ioctl_fd);
 #if 0	    /* HP-UX requires root privileges upon SIOC_IO ioctl */
@@ -1190,8 +1261,9 @@ char *setup_fds (char *device)
     {	case 0x11:	/* DVD-R			*/
 	case 0x13:	/* DVD-RW Restricted Overwrite	*/
 	case 0x14:	/* DVD-RW Sequential		*/
-	case 0x1B:	/* DVD+R			*/
 	case 0x1A:	/* DVD+RW			*/
+	case 0x1B:	/* DVD+R			*/
+	case 0x2B:	/* DVD+R Double Layer		*/
 	open_poor_man:
 	    poor_man = 1;
 	    out_fd = dup (ioctl_fd);
@@ -1235,7 +1307,7 @@ int setup_C_parm (char *C_parm,struct iso_primary_descriptor *descr)
 	next_session *= 16;
 	sprintf (C_parm,"16,%u",next_session);
     }
-    else if (profile==0x1B || profile==0x11 || profile==0x14)
+    else if (profile==0x2B || profile==0x1B || profile==0x11 || profile==0x14)
 	next_session=plusminus_r_C_parm (ioctl_handle,C_parm);
 
   return next_session;
@@ -1243,7 +1315,8 @@ int setup_C_parm (char *C_parm,struct iso_primary_descriptor *descr)
 
 int builtin_dd (int infd,int outfd,off64_t outoff)
 { char		*s;
-  int   	 n,off,fd;
+  int		 n,fd;
+  unsigned int	 off;
   struct stat64  sb;
   off64_t	 capacity=0,tracksize=0,startoff=outoff;
   pid_t		 pid=(pid_t)-1,ppid=getpid();
@@ -1253,7 +1326,7 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 	fprintf (stderr,":-( unable to mkstemp(\"%s\")",s),
 	exit(FATAL_START(errno));
 
-    ftruncate(fd,sizeof(*progress));
+    ftruncate(fd,(off_t)sizeof(*progress));
     unlink(s), free(s);
 
     progress = (void *)mmap (NULL,sizeof(*progress),PROT_READ|PROT_WRITE,
@@ -1299,6 +1372,7 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 	close (outfd);
 	while (kill (ppid,0)==0)
 	{   gettimeofday (&tv,NULL); slept =  tv.tv_usec/1e6+tv.tv_sec;
+	    lastcurrent = progress->current;
 	    poll (NULL,0,3333);
 	    gettimeofday (&tv,NULL); slept -= tv.tv_usec/1e6+tv.tv_sec;
 	    if (progress->zero==0 || !nfirst++) continue;
@@ -1343,9 +1417,6 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 					    tracksize/2048);
 			    if (overburn)
 				fprintf (stderr,":-! ignoring...\n");
-			    else if (dry_run)
-				close(infd), close(outfd),
-				exit(FATAL_START(ENOSPC));
 			    else
 			    {	n = -1; errno = FATAL_START(ENOSPC);
 				goto out;
@@ -1353,6 +1424,18 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 			}
 		    }
 		    /* else already checked for overburn condition */
+
+		    /* layer_break is meaningful only for -Z recording */
+		    if (layer_break>0 && !outoff)
+		    {	if (tracksize > layer_break*CD_BLOCK*2)
+			{   fprintf (stderr,":-( insane Layer Break position "
+					    "with respect to dataset size\n");
+			    n = -1; errno = FATAL_START(EINVAL);
+			    goto out;
+			}
+			if (!progress->final) progress->final = tracksize;
+			tracksize = layer_break*CD_BLOCK*2;
+		    }
 		}
 		else if (capacity > outoff)
 		{ int i=0;
@@ -1460,6 +1543,7 @@ int builtin_dd (int infd,int outfd,off64_t outoff)
 out:
     { int saved_errno=errno;
 
+	if (dry_run)	    close(infd), close(outfd), exit(errno);
 	if (pid!=(pid_t)-1) kill (pid,SIGTERM), waitpid (pid,NULL,0);
 
 	errno = outoff ? saved_errno : (n--?saved_errno:FATAL_START(EIO));
@@ -1533,9 +1617,11 @@ int main (int argc, char *argv[])
   int  i,n,warn_for_isofs=0;
   char **mkisofs_argv,C_parm[24],M_parm_[16],*M_parm=M_parm_;
   int  mkisofs_argc,growisofs_argc;
-  int  next_session=-1,alleged_next_session=-1,new_size;
+  int  next_session=-1,alleged_next_session=-1;
+  unsigned int new_size;
   struct iso_primary_descriptor *descr;
 
+#if !defined(I_KNOW_ALL_ABOUT_SUDO)
     if (getenv ("SUDO_COMMAND"))
     {	fprintf (stderr,":-( %s is being executed under sudo, "
 			"aborting!\n",argv[0]);
@@ -1543,6 +1629,7 @@ int main (int argc, char *argv[])
 			"manual page for further details.\n");
 	exit(FATAL_START(EACCES));
     }
+#endif
     /*
      * This is a set-root-uid "entry point" for listed operations. User
      * can't trick this code to unmount arbitrary file system, as [s]he
@@ -1578,7 +1665,7 @@ int main (int argc, char *argv[])
     /* mmap buffer so that we can use it with /dev/raw/rawN */
 #if defined(MAP_ANONYMOUS) && !(defined(__sun) || defined(sun))
     the_buffer = mmap (NULL,2*DVD_BLOCK,PROT_READ|PROT_WRITE,
-			MAP_PRIVATE|MAP_ANONYMOUS,-1,0);
+			MAP_PRIVATE|MAP_ANONYMOUS,-1,(off_t)0);
 #else
     { int fd;
 
@@ -1673,6 +1760,9 @@ int main (int argc, char *argv[])
 		    if (strstr(s,"tty"))	no_tty_check = 1;
 		    if (strstr(s,"dummy"))	test_write   = 1;
 		    if (strstr(s,"notray"))	no_reload    = 1;
+		    if (strstr(s,"noload"))	no_reload    = -1;
+		    if (strstr(s,"wrvfy"))	wrvfy        = 1;
+		    if (strstr(s,"4gms"))	no_4gb_check = 1;
 		    if (strstr(s,"moi"))
 		    {	quiet=-1; mkisofs_argv[mkisofs_argc++] = "-quiet";   }
 		    if ((o=strstr(s,"dao")))
@@ -1680,7 +1770,7 @@ int main (int argc, char *argv[])
 			/*  vvvvvvvvvvv tracksize option takes precedence! */
 			if (dao_size==0 && (o[3]==':' || o[3]=='='))
 			{   dao_size=strtol(o+4,0,0);
-			    if (dao_size%16)
+			    if (dao_size<=0)
 				fprintf (stderr,":-( insane dao%c%d option\n",
 						o[3],dao_size),
 				exit(FATAL_START(EINVAL));
@@ -1689,16 +1779,25 @@ int main (int argc, char *argv[])
 		    if ((o=strstr(s,"tracksize")))
 		    {	if (o[9]==':' || o[9]=='=')
 			{   dao_size=strtol(o+10,0,0);
-			    if (dao_size%16)
+			    if (dao_size<=0)
 				fprintf (stderr,":-( insane tracksize%c%d option\n",
 						o[9],dao_size),
 				exit(FATAL_START(EINVAL));
 			}
 		    }
+		    if ((o=strstr(s,"break")))
+		    {	if (o[5]==':' || o[5]=='=')
+			{   layer_break=strtol(o+6,0,0);
+			    if (layer_break<=0 || layer_break%16)
+				fprintf (stderr,":-( insane break%c%d option\n",
+						o[5],layer_break),
+				exit(FATAL_START(EINVAL));
+			}
+		    }
 		    if ((o=strstr(s,"seek")) && next_session<0)
 		    {	if (o[4]==':' || o[4]=='=')
 			{   next_session=strtol(o+5,0,0);
-			    if (next_session%16)
+			    if (next_session<0 || next_session%16)
 				fprintf (stderr,":-( insane seek%c%d option\n",
 						o[4],next_session),
 				exit(FATAL_START(EINVAL));
@@ -1729,7 +1828,7 @@ int main (int argc, char *argv[])
 	    else if (argv[i][1] == '?' || !strcmp(opt,"-help"))
 	    {	PRINT_VERSION (argv[0]);
 		printf ("- usage: %s [-dvd-compat] [-overburn] [-speed=1] \\\n"
-			"         -[ZM] /dev/cdrom <mkisofs options>\n",argv[0]);
+			"         -[ZM] /dev/dvd <mkisofs options>\n",argv[0]);
 		printf ("  for <mkisofs options> see 'mkisofs %s'\n",opt);
 		exit (FATAL_START(EINVAL));
 	    }
@@ -1833,9 +1932,10 @@ int main (int argc, char *argv[])
 			exit(FATAL_START(EINVAL));
 		}
 		else if (next_session > (0x200000-0x5000)) /* 4GB/2K-40MB/2K */
-		    fprintf (stderr,":-( next session would cross 4GB boundary, "
-				    "aborting...\n"),
-		    exit (FATAL_START(ENOSPC));
+		    if ((mmc_profile&0xFFFF)!=0x2B || !no_4gb_check)
+			fprintf (stderr,":-( next session would cross 4GB "
+					"boundary, aborting...\n"),
+			exit (FATAL_START(ENOSPC));
 
 		mkisofs_argv[mkisofs_argc++] = "-C";
 		mkisofs_argv[mkisofs_argc++] = C_parm;
@@ -1893,7 +1993,8 @@ int main (int argc, char *argv[])
     /* never finalize disc at multi-sessioning DVDR recordings...	*/
     { int profile = mmc_profile&0xFFFF;
 	if (next_session>0 &&
-	    (profile==0x1B || profile==0x11))	dvd_compat=0;
+	    (profile==0x2B || profile==0x1B || profile==0x11))
+						dvd_compat=0;
 	/* ... except when filling the media up:-)			*/
 	if (next_session>0 && zero_image)	dvd_compat=1;
     }
@@ -2003,6 +2104,7 @@ int main (int argc, char *argv[])
 	    case 0x14:	/* DVD-RW Sequential		*/
 	    case 0x1A:	/* DVD+RW			*/
 	    case 0x1B:	/* DVD+R			*/
+	    case 0x2B:	/* DVD+R Double Layer		*/
 		break;
 	    default:
 		fprintf (stderr,":-( %s: mounted media doesn't appear to be "
