@@ -43,7 +43,7 @@ static unsigned int  next_wr_addr=1;	// it starts as boolean
 static unsigned int  buf_size=256;	// measured in 1KBs
 static unsigned char formats[260],disc_info[32];
 
-extern int	dvd_compat,test_write,no_reload,mmc_profile,_argc;
+extern int	dvd_compat,test_write,no_reload,mmc_profile,_argc,wrvfy;
 extern double	speed_factor;
 extern char	*ioctl_device,**_argv;
 
@@ -76,7 +76,7 @@ int media_reload (char *name=NULL,struct stat *sb=NULL)
 #else
 #define RELOAD_NEVER_NEEDED 0
 #endif
-	if (RELOAD_NEVER_NEEDED || no_reload)
+	if (RELOAD_NEVER_NEEDED || no_reload>0)
 	{   cmd[0] = 0x1E;	// ALLOW MEDIA REMOVAL
 	    cmd[5] = 0;
 	    cmd.transport ();
@@ -132,6 +132,7 @@ int media_reload (char *name=NULL,struct stat *sb=NULL)
 #endif
 	    else return 0;	// m-m-m-m! patched kernel:-)
 	}
+	if (no_reload>=0)
 	{ Scsi_Command cmd;
 	    if (cmd.associate (name,sb))
 	    {	cmd[0] = 0x1B;		// START/STOP UNIT
@@ -208,7 +209,7 @@ int get_mmc_profile (void *fd)
 #endif
     } while (once--);
 
-    if (profile==0 || (profile&0x10)==0)	// no or non-DVD media...
+    if (profile==0 || (profile&0x30)==0)	// no or non-DVD media...
 	return profile;
 
     cmd[0] = 0x51;	// READ DISC INFORMATION
@@ -294,9 +295,10 @@ static unsigned int get_2k_capacity (Scsi_Command &cmd)
 	    else if (nwa<2048)	ret /= 2048/nwa;
 	    break;
 
-	case 0x1B:		// DVD+R
 	case 0x11:		// DVD-R
 	case 0x14:		// DVD-RW Sequential
+	case 0x1B:		// DVD+R
+	case 0x2B:		// DVD+R Double Layer
 	    cmd[0] = 0x52;	// READ TRACK INFORMATION
 	    cmd[1] = 1;
 	    cmd[4] = next_track>>8;
@@ -373,7 +375,7 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
     }
 
     while (1)
-    {	cmd[0] = 0x2A;			// WRITE(10)
+    {	cmd[0] = wrvfy?0x2E:0x2A;	// WRITE [AND VERIFY] (10)
 	cmd[2] = (lba>>24)&0xff;	// Logical Block Addrss
 	cmd[3] = (lba>>16)&0xff;
 	cmd[4] = (lba>>8)&0xff;
@@ -381,6 +383,17 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
 	cmd[7] = (nbl>>8)&0xff;
 	cmd[8] = nbl&0xff;
 	cmd[9] = 0;
+#if 0
+	cmd[0] = 0xAA;			// WRITE(12)
+	cmd[2] = (lba>>24)&0xff;	// Logical Block Addrss
+	cmd[3] = (lba>>16)&0xff;
+	cmd[4] = (lba>>8)&0xff;
+	cmd[5] = lba&0xff;
+	cmd[8] = (nbl>>8)&0xff;
+	cmd[9] = nbl&0xff;
+	cmd[10] = 0x80;			// "Streaming"
+	cmd[11] = 0;
+#endif
 	//
 	// First writes can be long, especially in DAO mode...
 	// I wish I could complement this with "if (lba==0),"
@@ -503,7 +516,8 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
 	    {	if (++retries > 1)
 	    	{   fprintf (stderr,":-! the LUN appears to be stuck at %xh, "
 		    		    "retrying in 5 secs...\n",lba),
-		    retries=0,
+		    retries=0;
+		    sigs_mask (SIG_UNBLOCK);
 		    poll (NULL,0,5000);
 		}
 		else if (errcode==0x52C00)
@@ -525,6 +539,9 @@ ssize_t poor_mans_pwrite64 (int fd,const void *_buff,size_t size,off64_t foff)
 				"-dvd-compat -dvd-compat to engage DAO or "
 				"apply full blanking procedure\n");
 	}
+#if 1	// safety net: leave DL sessions open and resumable...
+	if ((mmc_profile&0xFFFF)==0x2B && errcode!=0x52100) media_written=0;
+#endif
 	return -1;
     }
 
@@ -635,7 +652,8 @@ int plusminus_r_C_parm (void *fd,char *C_parm)
 	fprintf (stderr,":-( media is not appendable\n"),
 	exit(FATAL_START(EMEDIUMTYPE));
 
-    if (((disc_info[2]>>2)&3) != 0)
+    // allow to resume even --v-- incomplete sessions //
+    if (((disc_info[2]>>2)&3) > 1)
 	fprintf (stderr,":-( last session is not empty\n"),
 	exit(FATAL_START(EMEDIUMTYPE));
 
@@ -745,10 +763,13 @@ static int pull_velocity (Scsi_Command &cmd,unsigned char *d)
   int v,err;
   class autofree perf;
 
-    if ((d[4]&2) == 0)	return -1;	// LG GCA-4040N:-(
+#if 0	// 8x AccessTek derivatives, such as OptoRite DD0405
+    if ((d[4]&2) == 0)	return -1;
+#endif
 
     len = (d[0]<<24|d[1]<<16|d[2]<<8|d[3])-4;
     if (len%16)		return -1;
+    if (len==0)		return -1;	// LG GCA-4040N:-(
 
     len += 8;
     if (len == (8+16))
@@ -1180,9 +1201,19 @@ opc:
 	cmd[1] = 1;	// "Perform OPC"
 	cmd[9] = 0;
 	cmd.timeout(120);	// NEC units can be slooo...w
-	if ((err=cmd.transport()))
+	if ((err=cmd.transport())) do
+	{   if (err==0x52000)	// "INVALID COMMAND"
+		break;		// LD-F321 doesn't support SEND OPC?
+	    if (err==0x52700)	// "WRITE PROTECTED"
+		break;		// Plextor PX-712A?
+	    if (err==0x17301)	// "POWER CALIBRATION AREA ALMOST FULL"
+	    {	fprintf (stderr,":-! WARNING: Power Calibration Area "
+	    			    "is almost full\n");
+		break;
+	    }
 	    sperror ("PERFORM OPC",err),
 	    exit(FATAL_START(errno));
+	} while (0);
     }
 
     handle_events(cmd);
@@ -1278,6 +1309,63 @@ static int minus_r_reserve_track (Scsi_Command &cmd,off64_t size)
   return 0;
 }
 
+static void plus_r_dl_split (Scsi_Command &cmd,off64_t size)
+{ int           err;
+  unsigned int  blocks,split;
+  unsigned char dvd_20[4+8];
+
+    cmd[0] = 0xAD;	// READ DVD STRUCTURE
+    cmd[7] = 0x20;	// "DVD+R Double Layer Boundary Information"
+    cmd[9] = sizeof(dvd_20);
+    cmd[11] = 0;
+    if ((err=cmd.transport(READ,dvd_20,sizeof(dvd_20))))
+    	sperror ("READ DVD+R DL BOUNDARY INFORMATION",err),
+	exit (FATAL_START(errno));
+
+    if ((dvd_20[0]<<8|dvd_20[1]) < 10) 
+	fprintf (stderr,":-( insane DVD+R DL BI structure length\n"),
+	exit (FATAL_START(EINVAL));
+
+    if (dvd_20[4]&0x80)
+    {	fprintf (stderr,":-? L0 Data Zone Capacity is set already\n");
+	return;
+    }
+
+    split = dvd_20[8]<<24|dvd_20[9]<<16|dvd_20[10]<<8|dvd_20[11];
+
+    blocks = size/2048;
+    blocks += 15, blocks &= ~15;
+
+    if (blocks <= split)
+	fprintf (stderr,":-( more than 50%% of space will be *wasted*!\n"
+			"    use single layer media for this recording\n"),
+	exit (FATAL_START(EMEDIUMTYPE));
+
+    blocks /= 16;
+    blocks += 1;
+    blocks /= 2;
+    blocks *= 16;
+
+    fprintf (stderr,"%s: splitting layers at %u blocks\n",
+		    ioctl_device,blocks);
+
+    memset (dvd_20,0,sizeof(dvd_20));
+    dvd_20[1]  = sizeof(dvd_20)-2;
+    dvd_20[8]  = blocks>>24;
+    dvd_20[9]  = blocks>>16;
+    dvd_20[10] = blocks>>8;
+    dvd_20[11] = blocks;
+
+    cmd[0] = 0xBF;	// SEND DVD STRUCTURE
+    cmd[7] = 0x20;	// "DVD+R Double Layer Recording Information"
+    cmd[9] = sizeof(dvd_20);
+    cmd[11] = 0;
+    if ((err=cmd.transport(WRITE,dvd_20,sizeof(dvd_20))))
+	sperror ("SEND DVD+R DOUBLE LAYER RECORDING INFORMATION",err),
+	exit (FATAL_START(errno));
+
+}
+
 static int flush_cache (Scsi_Command &cmd)
 { int err;
 
@@ -1384,9 +1472,13 @@ void plus_r_finalize ()
 	if (wait_for_unit (cmd)) break;
 
 	if (dvd_compat)
-	    mode=0x06, fprintf (stderr,"%s: closing disc\n",ioctl_device);
+	    // HP dvd520n insists on "finalize with minimum radius"
+	    // with DVD+R DL ---------------vv
+	    mode = ((mmc_profile&0xFFFF)==0x2B)?0x05:0x06,
+	    fprintf (stderr,"%s: closing disc\n",ioctl_device);
 	else
-	    mode=0x02, fprintf (stderr,"%s: closing session\n",ioctl_device);
+	    mode = 0x02,
+	    fprintf (stderr,"%s: closing session\n",ioctl_device);
 
 	cmd[0] = 0x5B;		// CLOSE TRACK/SESSION
 	cmd[1] = 0x01;		// "IMMED"
@@ -1487,6 +1579,13 @@ void minus_r_finalize ()
     _exit (errno);
 }
 
+extern "C"
+void ram_reload ()
+{   sigs_mask(SIG_BLOCK);
+    if (media_written && do_reload) media_reload();
+    _exit (errno);
+}
+
 //
 // poor_mans_setup takes care of a lot of things.
 // It's invoked right before first write and if necessary/applicable
@@ -1551,7 +1650,10 @@ pwrite64_t poor_mans_setup (void *fd,off64_t leadout)
 	    atsignals (plus_rw_finalize);
 	    break;
 	case 0x1B:	// DVD+R
+	case 0x2B:	// DVD+R Double Layer
 	    plusminus_pages_setup(cmd,profile);
+	    if (profile==0x2B && next_track==1 && dvd_compat && leadout)
+		plus_r_dl_split (cmd,leadout);
 	    atexit (plus_r_finalize);
 	    if (next_wr_addr)
 	    {	atsignals (no_r_finalize);
@@ -1584,6 +1686,8 @@ pwrite64_t poor_mans_setup (void *fd,off64_t leadout)
 	case 0x12:	// DVD-RAM
 	    // exclusively for speed settings...
 	    plusminus_pages_setup (cmd,profile);
+	    atexit (ram_reload);
+	    atsignals (ram_reload);
 	    break;
 	default:
 	    fprintf (stderr,":-( mounted media[%X] is not supported\n",profile),
